#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/of_platform.h>
#include <linux/leds.h>
#include <linux/timer.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/of_platform.h>
#include <asm/gpio.h>

#include <linux/echocpu/pci_fpga.h>

#include <linux/mmc/host.h>
#include "../mmc/host/sdhci.h"

#define DRIVER_VERSION     "0.1"

#define PCI_FPGA_GPIO_REG          0x48
#define PCI_FPGA_LED_REG           0x4C
#define PCI_FPGA_ISA_IRQ_MASK      0x50
#define PCI_FPGA_ISA_IRQ_STATUS    0x54
#define PCI_FPGA_ISA_IRQ_TYPE      0x5c
#define PCI_FPGA_STATUS_REG        0x58
#define PCI_FPGA_STATUS_REG_SD_CD  (1 << 29)

#define PCI_VENDOR_ECHOCPU         0x10ee
#define PCI_DEVICE_ECHOCPU_FPGA    0x0003

#define ECHOCPU_NUM_LEDS           11
#define ECHOCPU_GPIO_COUNT         16

#define LED_NAME_SIZE 16

/* Delay must be at least 2 jiffies */
#define BOOT_LED_DELAY (HZ)

struct pci_fpga_drv_data;

struct pci_fpga_led_ctrl {
	unsigned int index;
	struct pci_fpga_drv_data *fpga_dev;
	struct led_classdev cdev;
};

struct pci_fpga_drv_data {
	void __iomem *regs;
	int bars;
	atomic_t do_leds;
	struct timer_list boot_led_timer;
	struct gpio_chip gpios;
	struct pci_fpga_led_ctrl leds[ECHOCPU_NUM_LEDS];
	struct sdhci_host* sdhost;
};

static const char *led_name[ECHOCPU_NUM_LEDS] = {
	"not_visible",
	"mount",
	"yellow_upper",
	"yellow_lower",
	"red_upper",
	"red_lower",
	"green_left_upper",
	"green_left_lower",
	"green_right_upper",
	"green_right_lower",
	"power"
};
static const char *pci_fpga_gpio_label = "pci_fpgag_pio";

static char pci_fpga_driver_name[] = "echocpu_fpga";

struct pci_fpga_drv_data *glb_drv_data = NULL;

static struct device_node *fpga_pic_node;
static struct irq_host *fpga_pic_host;

static struct pci_device_id pci_fpga_pci_tbl[] = {
	{ PCI_DEVICE(PCI_VENDOR_ECHOCPU, PCI_DEVICE_ECHOCPU_FPGA) },
	{ 0, }
};

static void do_boot_led_timer(unsigned long data)
{
	struct pci_fpga_drv_data *drv_data;
	u32 reg;

	drv_data = (struct pci_fpga_drv_data *)data;

	reg = in_le32(drv_data->regs + PCI_FPGA_LED_REG);

	if ((reg & 0x40) != 0) {
		reg &= ~0x40;
		reg |= 0x0100;
	} else if ((reg & 0x0100) != 0) {
		reg &= ~0x0100;
		reg |= 0x0200;
	} else if ((reg & 0x0200) != 0) {
		reg &= ~0x0200;
		reg |= 0x80;
	} else {
		reg &= ~0x80;
		reg |= 0x40;
	}

	/* Reschedule the timer */
	if (atomic_read(&drv_data->do_leds) != 0) {
		drv_data->boot_led_timer.expires = jiffies + BOOT_LED_DELAY;
		add_timer(&drv_data->boot_led_timer);
	} else {
		reg &= ~0x3C0;
	}
	out_le32(drv_data->regs + PCI_FPGA_LED_REG, reg);
}

int pci_fpga_get_card_detect( struct device *dev )
{
	struct pci_fpga_drv_data *drv_data;

	drv_data = glb_drv_data;
	
	if (drv_data == NULL) {
		return 0;
	}

	if ((in_le32(drv_data->regs + PCI_FPGA_STATUS_REG) & PCI_FPGA_STATUS_REG_SD_CD) != 0) {
		return 1;
	}
	
	return 0;
}
EXPORT_SYMBOL(pci_fpga_get_card_detect);

/* device open method */
static int pci_fpga_open(struct inode *inode, struct file *file)
{
	file->private_data = glb_drv_data;
	return(0);
}

/* device close method */
static int pci_fpga_release(struct inode *inode, struct file *file)
{
	return(0);
}

static ssize_t pci_fpga_write(struct file *file, const char *buf, size_t len, loff_t *off)
{
	struct pci_fpga_drv_data *drv_data;
	ssize_t ret;

	drv_data = (struct pci_fpga_drv_data *)file->private_data;
	ret = len;

	if (len < 2) {
		return -EINVAL;
	}

	if (len >= 2 && strncmp(buf, "on", 2) == 0) {
		atomic_set(&drv_data->do_leds, 1);
		drv_data->boot_led_timer.expires = jiffies + BOOT_LED_DELAY;
		add_timer(&drv_data->boot_led_timer);
	} else if (len >= 3 && strncmp(buf, "off", 3) == 0) {
		atomic_set(&drv_data->do_leds, 0);
	} else {
		ret = -EINVAL;
	}

	return ret;
}

static int pci_fpga_ioctl(struct inode *i, struct file *file, unsigned int command, unsigned long param)
{
	u32 value;
	int ret;
	pci_fpga_set_reg_t reg_req;
	struct pci_fpga_drv_data *drv_data;

	drv_data = file->private_data;
	
	ret = 0;
	switch (command) {
		case PCI_FPGA_MODIFY_REG:
			if (copy_from_user(&reg_req, (void *)param, sizeof(pci_fpga_set_reg_t)) != 0) {
				ret = -EFAULT;
				goto ioctl_out;
			}
			
			value = in_le32(drv_data->regs + reg_req.reg_addr);
			value &= ~reg_req.clearmask;
			value |= reg_req.setmask;
			out_le32(drv_data->regs + reg_req.reg_addr, value);
			break;
		case PCI_FPGA_READ_REG:
			if (copy_from_user(&reg_req, (void *)param, sizeof(pci_fpga_set_reg_t)) != 0) {
				ret = -EFAULT;
				goto ioctl_out;
			}
			
			reg_req.setmask = in_le32(drv_data->regs + reg_req.reg_addr);
			
			if (copy_to_user((void *)param, &reg_req, sizeof(pci_fpga_set_reg_t)) != 0) {
				ret = -EFAULT;
				goto ioctl_out;
			}
			break;
		default:
			ret = -ENOSYS;
	}

ioctl_out:
	return ret;
}

/* the ordinary device operations */
static struct file_operations pci_fpga_fops =
{
	owner:		THIS_MODULE,
	ioctl:		pci_fpga_ioctl,
	open:		pci_fpga_open,
	write:		pci_fpga_write,
	release:	pci_fpga_release
};

static struct miscdevice pci_fpga_misc = { MISC_DYNAMIC_MINOR, "pci_fpga", &pci_fpga_fops };

/* >32 here means not available */
#define ISA_IRQ_COUNT  16
static const unsigned int fpga_irq_map[ISA_IRQ_COUNT] = {
	33, 33, 33, 1, 2, 3, 4, 5,
	33, 6, 7, 8, 9, 33, 10, 11 };
	
static unsigned int fpga_irq_hw_to_virt_map[ISA_IRQ_COUNT];

static void pci_fpga_mask_irq(unsigned int irq)
{
	u32 value;
	unsigned int hwirq;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = glb_drv_data;
	hwirq = virq_to_hw(irq);
	
	value = in_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK);
	value &= ~(1 << fpga_irq_map[hwirq]);
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK, value);
}

static void pci_fpga_mask_ack_irq(unsigned int irq)
{
	u32 value;
	unsigned int hwirq;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = glb_drv_data;
	hwirq = virq_to_hw(irq);
	
	value = in_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK);
	value &= ~(1 << fpga_irq_map[hwirq]);
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK, value);
	
	value = 1 << fpga_irq_map[hwirq];
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_STATUS, value);
}

static void pci_fpga_ack_irq(unsigned int irq)
{
	u32 value;
	unsigned int hwirq;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = glb_drv_data;
	hwirq = virq_to_hw(irq);
	
	value = 1 << fpga_irq_map[hwirq];
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_STATUS, value);
}

static void pci_fpga_unmask_irq(unsigned int irq)
{
	u32 value;
	unsigned int hwirq;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = glb_drv_data;
	hwirq = virq_to_hw(irq);
	
	value = in_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK);
	value |= (1 << fpga_irq_map[hwirq]);
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_MASK, value);
}

static int pci_fpga_set_type(unsigned int, unsigned int);

static struct irq_chip pci_fpga_pic = {
	.typename = "FPGA",
	.mask = pci_fpga_mask_irq,
	/* For edge irq an ack is needed */
	.ack = pci_fpga_ack_irq,
	/* For level interrupt no ack is needed, just mask it */
	.mask_ack = pci_fpga_mask_ack_irq,
	.unmask = pci_fpga_unmask_irq,
	.set_type = pci_fpga_set_type,
};

static int pci_fpga_set_type(unsigned int virq, unsigned int flags)
{
	struct pci_fpga_drv_data *drv_data;
	u32 value;
	unsigned int hwirq;

	drv_data = glb_drv_data;
	hwirq = virq_to_hw(virq);

	if ((flags & (IRQ_TYPE_LEVEL_LOW | IRQ_TYPE_EDGE_FALLING)) != 0) {
		printk(KERN_ERR "pci_fpga: ISA interrupts can only be level high or rising edge\n");
		return -ENODEV;
	}
	value = in_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_TYPE);
	if ((flags & IRQ_TYPE_LEVEL_HIGH) != 0) {
		printk("IRQ %d registered as level high\n", hwirq);
		value |= (1 << fpga_irq_map[hwirq]);
		set_irq_handler(virq, handle_level_irq);
	} else {
		printk("IRQ %d registered as trigger rising\n", hwirq);
		value &= ~(1 << fpga_irq_map[hwirq]);
		set_irq_handler(virq, handle_edge_irq);
	}
	out_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_TYPE, value);
	
	return 0;
}

static int pci_fpga_pic_host_match(struct irq_host *h, struct device_node *node)
{
	return fpga_pic_node == node;
}

static int pci_fpga_pic_host_map(struct irq_host *h, unsigned int virq,
			     irq_hw_number_t hw)
{
	unsigned int hwirq;

	hwirq = virq_to_hw(virq);
	if (fpga_irq_map[hwirq] >= 32) {
		printk("fpga_pci: Not available ISA IRQ %d was requested. Only 3,4,5,6,7,9,10,11,12,14,15 are valid.\n", hwirq);
		return -1;
	}
	fpga_irq_hw_to_virt_map[hwirq] = virq;
	set_irq_chip(virq, &pci_fpga_pic);
	return 0;
}

static struct
irq_host_ops pci_fpga_pic_host_ops = {
	.match = pci_fpga_pic_host_match,
	.map = pci_fpga_pic_host_map,
};

static void pci_fpga_pic_cascade(unsigned int irq, struct irq_desc *desc)
{
	u32 value;
	struct pci_fpga_drv_data *drv_data;
	unsigned int i;

	drv_data = glb_drv_data;
	value = in_le32(drv_data->regs + PCI_FPGA_ISA_IRQ_STATUS);
	for (i=0; i<sizeof(fpga_irq_map) / sizeof(fpga_irq_map[0]); i++) {
		if (fpga_irq_map[i] < 32) {
			if ((value & (1 << fpga_irq_map[i])) != 0) {
				generic_handle_irq(fpga_irq_hw_to_virt_map[i]);
			}
		}
	}
}

static int pci_fpga_setup_irq_host(struct pci_dev *pdev)
{
	struct device_node *np;

	np = of_find_compatible_node(NULL, NULL, "fsl,echpcpu-fpga-pic");
	if (np == NULL) {
		printk(KERN_ERR "pci_fpga: Can not find fpga-pic node\n");
		return -ENODEV;
	}

	printk("Cascade IRQ is %d\n", pdev->irq);

	fpga_pic_node = of_node_get(np);

	fpga_pic_host = irq_alloc_host(np, IRQ_HOST_MAP_LINEAR, 16, &pci_fpga_pic_host_ops, 16);
	if (fpga_pic_host == NULL) {
		printk(KERN_ERR "pci_fpga: failed to allocate irq host!\n");
		goto end;
	}

	set_irq_chained_handler(pdev->irq, pci_fpga_pic_cascade);
end:
	of_node_put(np);
	
	return 0;
}

#define NODE_NAME_SIZE 32
static int pci_fpga_register_isa_devices(struct pci_dev *pdev)
{
	struct device_node *np;
	struct device_node *child;
	struct of_device *new_dev;
	struct device *parent;
	struct property *new;
	struct property *oldprop;
	u32 * values;
	resource_size_t start;
	int ret;
	
	parent = &pdev->dev;
	
	np = of_find_compatible_node(NULL, NULL, /*"fsl,echpcpu-fpga-pic"*/"fsl,mpc8349-pci");
	if (np == NULL) {
		printk(KERN_ERR "pci_fpga: Cannot find fpga-pic node\n");
		return -ENODEV;
	}
	
	for_each_child_of_node(np, child) {
		if (of_device_is_compatible(child, "isadev") != 0) {
			oldprop = of_find_property(child, "assigned-addresses",NULL);
			if (oldprop != NULL) {
				new = kzalloc(sizeof(struct property) + NODE_NAME_SIZE + (10 *  sizeof(u32)), GFP_KERNEL);
				if (new != NULL) {
					/* Put in the new adresses */
					new->name = (char *)(new + 1);
					new->value = new->name + NODE_NAME_SIZE;
					new->length = 10 *  sizeof(u32);
					values = new->value;
					strcpy(new->name, "assigned-addresses");
					start = pci_resource_start(pdev, 1);
					values[0] = 0x02000000;
					values[1] = 0x0;
					values[2] = start;
					values[3] = 0x0;
					values[4] = 0x10000;
					values[5] = 0x02000000;
					values[6] = 0x0;
					values[7] = start + 0x1000000;
					values[8] = 0x0;
					values[9] = 0x1000000;
					
					ret = prom_update_property(child, new, oldprop);
					if (ret != 0) {
						printk(KERN_ERR "pci_fpga: Unable to update property\n");
						continue;
					}
				}
			}
			printk("pci_fpga: Adding ISA device from OF\n");
			new_dev = of_platform_device_create(child, NULL, parent);
		}
	}
	
	return 0;
}

static void pci_fpga_set_led(struct led_classdev *led_cdev, enum led_brightness value)
{
	struct pci_fpga_led_ctrl *led_data =
		container_of(led_cdev, struct pci_fpga_led_ctrl, cdev);
	struct pci_fpga_drv_data *drv_data = led_data->fpga_dev;
	u32 reg_value;

	reg_value = in_le32(drv_data->regs + PCI_FPGA_LED_REG);
	if (value == LED_OFF) {
		reg_value &= ~(1 << led_data->index);
	} else {
		reg_value |= (1 << led_data->index);
	}
	out_le32(drv_data->regs + PCI_FPGA_LED_REG, reg_value);
}

static enum led_brightness pci_fpga_get_led(struct led_classdev *led_cdev)
{
	struct pci_fpga_led_ctrl *led_data =
		container_of(led_cdev, struct pci_fpga_led_ctrl, cdev);
	struct pci_fpga_drv_data *drv_data = led_data->fpga_dev;
	u32 reg_value;

	reg_value = in_le32(drv_data->regs + PCI_FPGA_LED_REG);
	
	if ((reg_value & (1 << led_data->index)) != 0) {
		return LED_FULL;
	}
	
	return LED_OFF;
}

static int pci_fpga_setup_leds(struct device *parent, struct pci_fpga_drv_data *drv_data)
{
	struct pci_fpga_led_ctrl *cur_led;
	unsigned int i;
	int ret;

	for (i = 0; i < ECHOCPU_NUM_LEDS; i++) {
		cur_led = &drv_data->leds[i];
		
		cur_led->index = i;
		cur_led->fpga_dev = drv_data;

		cur_led->cdev.name = led_name[i];
		cur_led->cdev.default_trigger = NULL;
		cur_led->cdev.brightness_set = pci_fpga_set_led;
		cur_led->cdev.brightness_get = pci_fpga_get_led;
		cur_led->cdev.default_trigger = NULL;
		cur_led->cdev.blink_set = NULL;
		cur_led->cdev.brightness = LED_OFF;

		ret = led_classdev_register(NULL, &cur_led->cdev);
		if (ret < 0) {
			printk("pci_fpga: Error registerring LED '%s'\n", cur_led->cdev.name);
		}
	}
	
	return 0;
}

static void pci_fpga_cleanup_leds(struct pci_fpga_drv_data *drv_data)
{
	unsigned int i;

	for (i = 0; i < ECHOCPU_NUM_LEDS; i++) {
		led_classdev_unregister(&drv_data->leds[i].cdev);
	}
}

static int pci_fpga_gpio_set_input(struct gpio_chip *chip, unsigned offset)
{
	u32 value;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = container_of(chip, struct pci_fpga_drv_data, gpios);
	
	value = in_le32(drv_data->regs + PCI_FPGA_GPIO_REG);
	value |= (1 << (offset + 16));
	out_le32(drv_data->regs + PCI_FPGA_GPIO_REG, value);
	
	return 0;
}

static int pci_fpga_gpio_set_output(struct gpio_chip *chip, unsigned offset, int new_value)
{
	u32 value;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = container_of(chip, struct pci_fpga_drv_data, gpios);
	
	value = in_le32(drv_data->regs + PCI_FPGA_GPIO_REG);
	if (new_value == 0) {
		value &= ~(1 << offset);
	} else {
		value |= (1 << offset);
	}
	value &= ~(1 << (offset + 16));
	out_le32(drv_data->regs + PCI_FPGA_GPIO_REG, value);
	
	return 0;
}

static int pci_fpga_gpio_get(struct gpio_chip *chip, unsigned offset)
{
	u32 value;
	struct pci_fpga_drv_data *drv_data;
	
	drv_data = container_of(chip, struct pci_fpga_drv_data, gpios);
	
	value = in_le32(drv_data->regs + PCI_FPGA_GPIO_REG);
	
	return ((value & (1 << offset)) != 0);
}

static void pci_fpga_gpio_set(struct gpio_chip *chip, unsigned offset, int new_value)
{
	u32 value;
	struct pci_fpga_drv_data *drv_data;

	drv_data = container_of(chip, struct pci_fpga_drv_data, gpios);

	value = in_le32(drv_data->regs + PCI_FPGA_GPIO_REG);
	if (new_value == 0) {
		value &= ~(1 << offset);
	} else {
		value |= (1 << offset);
	}
	out_le32(drv_data->regs + PCI_FPGA_GPIO_REG, value);
}

static int pci_fpga_setup_gpio(struct device *parent, struct pci_fpga_drv_data *drv_data)
{
	drv_data->gpios.base = -1;
	drv_data->gpios.can_sleep = 0;
	drv_data->gpios.dev = parent;
	drv_data->gpios.owner = THIS_MODULE;
	drv_data->gpios.ngpio = ECHOCPU_GPIO_COUNT;
	drv_data->gpios.direction_input = pci_fpga_gpio_set_input;
	drv_data->gpios.get = pci_fpga_gpio_get;
	drv_data->gpios.direction_output = pci_fpga_gpio_set_output;
	drv_data->gpios.set = pci_fpga_gpio_set;
	drv_data->gpios.label = pci_fpga_gpio_label;
	
	if (gpiochip_add(&drv_data->gpios) < 0) {
		printk(KERN_ERR "pci_fpga: Error registering GPIO pins of PCI FPGA\n");
	}
	
	return 0;
}

static int pci_fpga_cleanup_gpio(struct pci_fpga_drv_data *drv_data)
{
	if (gpiochip_remove(&drv_data->gpios) < 0) {
		printk(KERN_ERR "pci_fpga: Error releasing GPIO pins of PCI FPGA\n");
	}

	return 0;
}

static struct sdhci_ops sdhci_echocpu_ops = {
	.enable_dma = 0,
};

static int dummy_irq(int irq, void* dev_id);

static int __devinit pci_fpga_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct pci_fpga_drv_data *drv_data;
	int ret;

	/* Get mem for driver data */
	drv_data = kzalloc(sizeof(struct pci_fpga_drv_data), GFP_KERNEL);
	if (drv_data == NULL) {
		printk(KERN_ERR "pci_fpga: No memory for driver data\n");
		return -ENOMEM;
	}
	
	/* Get the BARS */
	drv_data->bars = pci_select_bars(pdev, IORESOURCE_MEM);
	/* Enable the device */
	ret = pci_enable_device_mem(pdev);
	
	if (ret != 0) {
		goto out_free;
	}

	/* Request the mem regions for bar access */
	ret = pci_request_selected_regions(pdev, drv_data->bars, pci_fpga_driver_name);
	if (ret != 0) {
		goto out_device;
	}

	drv_data->regs = pci_ioremap_bar(pdev, 0);
	if (drv_data->regs == NULL) {
		ret = -ENODEV;
		goto out_regions;
	}

	ret = misc_register(&pci_fpga_misc);
	if (ret != 0) {
		printk(KERN_ERR "pci_fpga: Registering misc device failed\n");
		ret = -ENODEV;
		goto out_ioremap;
	}
	printk("pci_fpga: Registered as %d, %d\n", MISC_MAJOR, pci_fpga_misc.minor);

	/* Remember the driver data */
	pci_set_drvdata(pdev, drv_data);
	glb_drv_data = drv_data;
	
	/* Setup the led drivers */
	pci_fpga_setup_leds(&pdev->dev, drv_data);
	
	/* Setup the GPIO driver */
	pci_fpga_setup_gpio(&pdev->dev, drv_data);
	
	/* Setup the IRQ host */
	pci_fpga_setup_irq_host(pdev);
	
	/* Register the ISA devices from device tree */
	pci_fpga_register_isa_devices(pdev);

	/* Start the timer */
	atomic_set(&drv_data->do_leds, 1);
	init_timer(&drv_data->boot_led_timer);
	drv_data->boot_led_timer.function = do_boot_led_timer;
	drv_data->boot_led_timer.data = (unsigned long)drv_data;
	drv_data->boot_led_timer.expires = jiffies + 2;
	add_timer(&drv_data->boot_led_timer);
	
	/* setup sd card driver */
#if 1
	printk(KERN_CRIT"init sdhost\n");
	drv_data->sdhost = sdhci_alloc_host(&pdev->dev, 4);
	if (drv_data->sdhost)
	{
		printk(KERN_CRIT"have sdhost; irq=%d\n", pdev->irq);
		drv_data->sdhost->hw_name = "echocpu_fpga";
		drv_data->sdhost->ops = &sdhci_echocpu_ops;
		drv_data->sdhost->quirks = 0; // TODO
		int virt_irq = irq_create_mapping (NULL, 74);
		/* HACK vvv */
		uint32_t* ptr = (uint32_t*)ioremap_nocache(0xE0000C00, 24);
		ptr[3]  = (0x80000000 >> 14); // no |=, reset interrupt state
		ptr[5] |= (0x80000000 >> 14);
		ptr[4] |= (0x80000000 >> 14);
		void* p = sdhci_priv(drv_data->sdhost);
		*(uint32_t**)p = ptr;
		/* HACK ^^^ */
		drv_data->sdhost->irq = virt_irq; // pdev->irq;
		ret = 0; // pci_request_region(pdev, 3, mmc_hostname(drv_data->sdhost->mmc)); // bar 3
		if (!ret)
		{
			printk(KERN_CRIT"have pci_region\n");
			drv_data->sdhost->ioaddr = pci_ioremap_bar(pdev, 3); // bar 3
			if (drv_data->sdhost->ioaddr)
			{
				printk(KERN_CRIT"have ioaddr\n");
				ret = sdhci_add_host(drv_data->sdhost);
				if (!ret)
				{
					printk(KERN_CRIT"sdhci_add_host ok\n");
				}
				else
				{
					printk(KERN_CRIT"sdhci_add_host failed: %d\n", ret);
				}
			}
		}
	}
	else
	{
		printk(KERN_CRIT"no sdhost\n");
	}
#endif
#if 0
	// dummy driver, acks interrupt but does nothing else
	int virt_irq = irq_create_mapping (NULL, 74);
	drv_data->sdhost = sdhci_alloc_host(&pdev->dev, 4);
	if (drv_data->sdhost)
	{
		void* p = sdhci_priv(drv_data->sdhost);
		uint32_t* ptr = (uint32_t*)ioremap_nocache(0xE0000C00, 24);
		ptr[3]  = (0x80000000 >> 14); // no |=, reset interrupt state
		ptr[5] |= (0x80000000 >> 14);
		ptr[4] |= (0x80000000 >> 14);
		*(uint32_t**)p = ptr;
		request_irq(virt_irq, dummy_irq, IRQF_SHARED, "dummy_driver", drv_data->sdhost);
		drv_data->sdhost->irq = virt_irq;
		drv_data->sdhost->ioaddr = pci_ioremap_bar(pdev, 3); // bar 3
		writel(0xFFFFFFFF, drv_data->sdhost->ioaddr + SDHCI_INT_ENABLE);
		writel(0xFFFFFFFF, drv_data->sdhost->ioaddr + SDHCI_SIGNAL_ENABLE);
	}
#endif

	return 0;

out_ioremap:
	iounmap(drv_data->regs);
out_regions:
	pci_release_selected_regions(pdev, drv_data->bars);
out_device:
	pci_disable_device(pdev);
out_free:
	kfree(drv_data);
	return ret;
}

#if 0
static int dummy_irq(int irq, void* dev_id)
{
	struct sdhci_host* host = (struct sdhci_host*)dev_id;
	uint32_t* p2 = *(uint32_t**)sdhci_priv(host);
	uint32_t intstatus = readl(host->ioaddr + SDHCI_INT_STATUS);
	printk(KERN_CRIT"sdhci_irq: jiffies %lu status %08X GPDAT14 %s GPIER14 %s\n", jiffies, intstatus, p2[2] & (0x80000000 >> 14) ? "ON" : "off", p2[3] & (0x80000000 >> 14) ? "ON" : "off");
	if (!intstatus)
	{
		printk(KERN_CRIT"sdhci_irq: not me\n");
		return IRQ_NONE;
	}
	// dummy, just ack the interrupt, do nothing
	writel(intstatus, host->ioaddr + SDHCI_INT_STATUS);
	uint32_t e = p2[3];
	p2[3] = (0x80000000 >> 14);
	printk(KERN_CRIT"sdhci_irq: handled status %08X GPDAT14 %s GPIER14 %s -> %s\n", readl(host->ioaddr + SDHCI_INT_STATUS), p2[2] & (0x80000000 >> 14) ? "ON" : "off", e & (0x80000000 >> 14) ? "ON" : "off", p2[3] & (0x80000000 >> 14) ? "ON" : "off");
	return IRQ_HANDLED;
}
#endif

static void __devexit pci_fpga_remove(struct pci_dev *pdev)
{
	struct pci_fpga_drv_data *drv_data;

	drv_data = pci_get_drvdata(pdev);
	
#if 1
	if (drv_data->sdhost)
	{
		int dead = 0;
		u32 scratch = readl(drv_data->sdhost->ioaddr + SDHCI_INT_STATUS);
		if (scratch == (u32)-1) dead = 1;
		sdhci_remove_host(drv_data->sdhost, dead);
		// pci_release_region(pdev, 3);
		void* p = sdhci_priv(drv_data->sdhost);
		iounmap(p);
		sdhci_free_host(drv_data->sdhost);
	}
#endif

	/* Stop led timer if pending */
	if (timer_pending(&drv_data->boot_led_timer)) {
		del_timer_sync(&drv_data->boot_led_timer);
	}

	pci_fpga_cleanup_leds(drv_data);

	pci_fpga_cleanup_gpio(drv_data);

	misc_deregister(&pci_fpga_misc);
	iounmap(drv_data->regs);
	pci_release_selected_regions(pdev, drv_data->bars);
	pci_disable_device(pdev);
	kfree(drv_data);
}

#ifdef CONFIG_PM
static int pci_fpga_suspend(struct pci_dev *pdev, pm_message_t state)
{
	return -ENOSYS;
}

static int pci_fpga_resume(struct pci_dev *pdev)
{
	return -ENOSYS;
}
#endif

static void pci_fpga_shutdown(struct pci_dev *pdev)
{
	return;
}

static pci_ers_result_t pci_fpga_io_error_detected(struct pci_dev *pdev, pci_channel_state_t state)
{
	return PCI_ERS_RESULT_NEED_RESET;
}

static pci_ers_result_t pci_fpga_io_slot_reset(struct pci_dev *pdev)
{
	struct pci_fpga_drv_data *drv_data;
	int ret;
	
	drv_data = pci_get_drvdata(pdev);

	ret = pci_enable_device_mem(pdev);
	if (ret != 0) {
		printk(KERN_ERR "pci_fpga: Failed to reenable PCI device after reset\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}

	return PCI_ERS_RESULT_RECOVERED;
}

static void pci_fpga_io_resume(struct pci_dev *pdev)
{
	return;
}

static struct pci_error_handlers pci_fpga_err_handler = {
	.error_detected = pci_fpga_io_error_detected,
	.slot_reset = pci_fpga_io_slot_reset,
	.resume = pci_fpga_io_resume,
};

static struct pci_driver echocpu_fpga_driver = {
	.name     = pci_fpga_driver_name,
	.id_table = pci_fpga_pci_tbl,
	.probe    = pci_fpga_probe,
	.remove   = __devexit_p(pci_fpga_remove),
#ifdef CONFIG_PM
	/* Power Managment Hooks */
	.suspend  = pci_fpga_suspend,
	.resume   = pci_fpga_resume,
#endif
	.shutdown = pci_fpga_shutdown,
	.err_handler = &pci_fpga_err_handler,
};

static int __init pci_fpga_init(void)
{
	int ret;
	printk(KERN_INFO "EchoCPU FPGA driver - version %s\n", DRIVER_VERSION);
	
	ret = pci_register_driver(&echocpu_fpga_driver);
	
	return ret;
}
module_init(pci_fpga_init);

static void __exit pci_fpga_exit(void)
{
	pci_unregister_driver(&echocpu_fpga_driver);
}

module_exit(pci_fpga_exit);
